/*
	LablScintilla - Scintilla [LablGTK] binding for Objective Caml

	author: Robert Roessler
	email: robertr@rftp.com
	web: http://www.rftp.com

--------------------------------------------------------------------------------
Copyright (c) 2005-2007 Robert Roessler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Robert Roessler nor the names of its contributors
      may be used to endorse or promote products derived from this
      software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
--------------------------------------------------------------------------------
*/

/* ml_scintilla.cpp - interface from Scintilla OCaml code -> C-land */

#include <gtk/gtk.h>
#include <string.h>

#include <Platform.h>
#if PLAT_GTK_WIN32
#include <Windows.h>
#endif
#include <Scintilla.h>
#include <ScintillaWidget.h>

extern "C" {
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/fail.h>

//	uncomment following to remove dependency on LablGTK sources (on Windows)
//#define REALLY_LATE_BINDING

#if defined(CAML_DLL) && defined(_WINDOWS_) && defined(REALLY_LATE_BINDING)
#define Pointer_val(val) ((void*)Field(val,1))
#define check_cast(f,v) f(Pointer_val(v))
#else
#include <wrappers.h>
#include <ml_gtk.h>
#include <ml_gobject.h>
#endif	/* CAML_DLL && _WINDOWS_ && REALLY_LATE_BINDING */

#define ScintillaObject_val(val) check_cast(SCINTILLA, val)

#define Val_int32 copy_int32

#define SSM(wP, lP) \
scintilla_send_message(ScintillaObject_val(obj), sci_cmd[Int_val(msg)], (wP), (lP))

#define SS_2(name, obj, msg, conv1, conv2, conv) \
CAMLprim value ml_sci_send_message_##name(value obj, value msg, value wP, value lP) \
{return conv(SSM(conv1(wP), conv2(lP)));}

static void prepend_flags_list(value& agg, int slot, unsigned int m);

//	Scintilla commands
static const int sci_cmd[] = {
	// ("standard" Scintilla commands)
	SCI_ADDREFDOCUMENT, SCI_ADDSTYLEDTEXT, SCI_ADDTEXT, SCI_ALLOCATE,
	SCI_APPENDTEXT, SCI_ASSIGNCMDKEY, SCI_AUTOCACTIVE, SCI_AUTOCCANCEL,
	SCI_AUTOCCOMPLETE, SCI_AUTOCGETAUTOHIDE, SCI_AUTOCGETCANCELATSTART,
	SCI_AUTOCGETCHOOSESINGLE, SCI_AUTOCGETCURRENT, SCI_AUTOCGETDROPRESTOFWORD,
	SCI_AUTOCGETIGNORECASE, SCI_AUTOCGETMAXHEIGHT, SCI_AUTOCGETMAXWIDTH,
	SCI_AUTOCGETSEPARATOR, SCI_AUTOCGETTYPESEPARATOR, SCI_AUTOCPOSSTART,
	SCI_AUTOCSELECT, SCI_AUTOCSETAUTOHIDE, SCI_AUTOCSETCANCELATSTART,
	SCI_AUTOCSETCHOOSESINGLE, SCI_AUTOCSETDROPRESTOFWORD, SCI_AUTOCSETFILLUPS,
	SCI_AUTOCSETIGNORECASE, SCI_AUTOCSETMAXHEIGHT, SCI_AUTOCSETMAXWIDTH,
	SCI_AUTOCSETSEPARATOR, SCI_AUTOCSETTYPESEPARATOR, SCI_AUTOCSHOW,
	SCI_AUTOCSTOPS, SCI_BEGINUNDOACTION, SCI_BRACEBADLIGHT, SCI_BRACEHIGHLIGHT,
	SCI_BRACEMATCH, SCI_CALLTIPACTIVE, SCI_CALLTIPCANCEL, SCI_CALLTIPPOSSTART,
	SCI_CALLTIPSETBACK, SCI_CALLTIPSETFORE, SCI_CALLTIPSETFOREHLT,
	SCI_CALLTIPSETHLT, SCI_CALLTIPSHOW, SCI_CALLTIPUSESTYLE, SCI_CANPASTE,
	SCI_CANREDO, SCI_CANUNDO, SCI_CHOOSECARETX, SCI_CLEAR, SCI_CLEARALL,
	SCI_CLEARALLCMDKEYS, SCI_CLEARCMDKEY, SCI_CLEARDOCUMENTSTYLE,
	SCI_CLEARREGISTEREDIMAGES, SCI_COLOURISE, SCI_CONVERTEOLS, SCI_COPY,
	SCI_COPYRANGE, SCI_COPYTEXT, SCI_CREATEDOCUMENT, SCI_CUT,
	SCI_DOCLINEFROMVISIBLE, SCI_EMPTYUNDOBUFFER, SCI_ENCODEDFROMUTF8,
	SCI_ENDUNDOACTION, SCI_ENSUREVISIBLE, SCI_ENSUREVISIBLEENFORCEPOLICY,
	SCI_FINDCOLUMN, SCI_FINDTEXT, SCI_FORMATRANGE, SCI_GETANCHOR,
	SCI_GETBACKSPACEUNINDENTS, SCI_GETBUFFEREDDRAW, SCI_GETCARETFORE,
	SCI_GETCARETLINEBACK, SCI_GETCARETLINEBACKALPHA, SCI_GETCARETLINEVISIBLE,
	SCI_GETCARETPERIOD, SCI_GETCARETSTICKY, SCI_GETCARETWIDTH, SCI_GETCHARAT,
	SCI_GETCODEPAGE, SCI_GETCOLUMN, SCI_GETCONTROLCHARSYMBOL, SCI_GETCURLINE,
	SCI_GETCURRENTPOS, SCI_GETCURSOR, SCI_GETDIRECTFUNCTION, SCI_GETDIRECTPOINTER,
	SCI_GETDOCPOINTER, SCI_GETEDGECOLOUR, SCI_GETEDGECOLUMN, SCI_GETEDGEMODE,
	SCI_GETENDATLASTLINE, SCI_GETENDSTYLED, SCI_GETEOLMODE,
	SCI_GETFIRSTVISIBLELINE, SCI_GETFOCUS, SCI_GETFOLDEXPANDED, SCI_GETFOLDLEVEL,
	SCI_GETFOLDPARENT, SCI_GETHIGHLIGHTGUIDE, SCI_GETHSCROLLBAR, SCI_GETINDENT,
	SCI_GETINDENTATIONGUIDES, SCI_GETLASTCHILD, SCI_GETLAYOUTCACHE, SCI_GETLENGTH,
	SCI_GETLEXER, SCI_GETLINE, SCI_GETLINECOUNT, SCI_GETLINEENDPOSITION,
	SCI_GETLINEINDENTATION, SCI_GETLINEINDENTPOSITION, SCI_GETLINESELENDPOSITION,
	SCI_GETLINESELSTARTPOSITION, SCI_GETLINESTATE, SCI_GETLINEVISIBLE,
	SCI_GETMARGINLEFT, SCI_GETMARGINMASKN, SCI_GETMARGINRIGHT,
	SCI_GETMARGINSENSITIVEN, SCI_GETMARGINTYPEN, SCI_GETMARGINWIDTHN,
	SCI_GETMAXLINESTATE, SCI_GETMODEVENTMASK, SCI_GETMODIFY,
	SCI_GETMOUSEDOWNCAPTURES, SCI_GETMOUSEDWELLTIME, SCI_GETOVERTYPE,
	SCI_GETPASTECONVERTENDINGS, SCI_GETPRINTCOLOURMODE, SCI_GETPRINTMAGNIFICATION,
	SCI_GETPRINTWRAPMODE, SCI_GETPROPERTY, SCI_GETPROPERTYEXPANDED,
	SCI_GETPROPERTYINT, SCI_GETREADONLY, SCI_GETSCROLLWIDTH, SCI_GETSEARCHFLAGS,
	SCI_GETSELALPHA, SCI_GETSELECTIONEND, SCI_GETSELECTIONMODE,
	SCI_GETSELECTIONSTART, SCI_GETSELTEXT, SCI_GETSTATUS, SCI_GETSTYLEAT,
	SCI_GETSTYLEBITS, SCI_GETSTYLEBITSNEEDED, SCI_GETSTYLEDTEXT,
	SCI_GETTABINDENTS, SCI_GETTABWIDTH, SCI_GETTARGETEND, SCI_GETTARGETSTART,
	SCI_GETTEXT, SCI_GETTEXTLENGTH, SCI_GETTEXTRANGE, SCI_GETTWOPHASEDRAW,
	SCI_GETUNDOCOLLECTION, SCI_GETUSEPALETTE, SCI_GETUSETABS, SCI_GETVIEWEOL,
	SCI_GETVIEWWS, SCI_GETVSCROLLBAR, SCI_GETWRAPMODE, SCI_GETWRAPSTARTINDENT,
	SCI_GETWRAPVISUALFLAGS, SCI_GETWRAPVISUALFLAGSLOCATION, SCI_GETXOFFSET,
	SCI_GETZOOM, SCI_GOTOLINE, SCI_GOTOPOS, SCI_GRABFOCUS, SCI_HIDELINES,
	SCI_HIDESELECTION, SCI_INDICGETFORE, SCI_INDICGETSTYLE, SCI_INDICSETFORE,
	SCI_INDICSETSTYLE, SCI_INSERTTEXT, SCI_LINEFROMPOSITION, SCI_LINELENGTH,
	SCI_LINESCROLL, SCI_LINESJOIN, SCI_LINESONSCREEN, SCI_LINESSPLIT,
	SCI_LOADLEXERLIBRARY, SCI_MARKERADD, SCI_MARKERADDSET, SCI_MARKERDEFINE,
	SCI_MARKERDEFINEPIXMAP, SCI_MARKERDELETE, SCI_MARKERDELETEALL,
	SCI_MARKERDELETEHANDLE, SCI_MARKERGET, SCI_MARKERLINEFROMHANDLE,
	SCI_MARKERNEXT, SCI_MARKERPREVIOUS, SCI_MARKERSETALPHA, SCI_MARKERSETBACK,
	SCI_MARKERSETFORE, SCI_MOVECARETINSIDEVIEW, SCI_NULL, SCI_PASTE,
	SCI_POINTXFROMPOSITION, SCI_POINTYFROMPOSITION, SCI_POSITIONAFTER,
	SCI_POSITIONBEFORE, SCI_POSITIONFROMLINE, SCI_POSITIONFROMPOINT,
	SCI_POSITIONFROMPOINTCLOSE, SCI_REDO, SCI_REGISTERIMAGE, SCI_RELEASEDOCUMENT,
	SCI_REPLACESEL, SCI_REPLACETARGET, SCI_REPLACETARGETRE, SCI_SCROLLCARET,
	SCI_SEARCHANCHOR, SCI_SEARCHINTARGET, SCI_SEARCHNEXT, SCI_SEARCHPREV,
	SCI_SELECTALL, SCI_SELECTIONISRECTANGLE, SCI_SETANCHOR,
	SCI_SETBACKSPACEUNINDENTS, SCI_SETBUFFEREDDRAW, SCI_SETCARETFORE,
	SCI_SETCARETLINEBACK, SCI_SETCARETLINEBACKALPHA, SCI_SETCARETLINEVISIBLE,
	SCI_SETCARETPERIOD, SCI_SETCARETSTICKY, SCI_SETCARETWIDTH,
	SCI_SETCHARSDEFAULT, SCI_SETCODEPAGE, SCI_SETCONTROLCHARSYMBOL,
	SCI_SETCURRENTPOS, SCI_SETCURSOR, SCI_SETDOCPOINTER, SCI_SETEDGECOLOUR,
	SCI_SETEDGECOLUMN, SCI_SETEDGEMODE, SCI_SETENDATLASTLINE, SCI_SETEOLMODE,
	SCI_SETFOCUS, SCI_SETFOLDEXPANDED, SCI_SETFOLDFLAGS, SCI_SETFOLDLEVEL,
	SCI_SETFOLDMARGINCOLOUR, SCI_SETFOLDMARGINHICOLOUR, SCI_SETHIGHLIGHTGUIDE,
	SCI_SETHOTSPOTACTIVEBACK, SCI_SETHOTSPOTACTIVEFORE,
	SCI_SETHOTSPOTACTIVEUNDERLINE, SCI_SETHOTSPOTSINGLELINE, SCI_SETHSCROLLBAR,
	SCI_SETINDENT, SCI_SETINDENTATIONGUIDES, SCI_SETKEYWORDS, SCI_SETLAYOUTCACHE,
	SCI_SETLENGTHFORENCODE, SCI_SETLEXER, SCI_SETLEXERLANGUAGE,
	SCI_SETLINEINDENTATION, SCI_SETLINESTATE, SCI_SETMARGINLEFT,
	SCI_SETMARGINMASKN, SCI_SETMARGINRIGHT, SCI_SETMARGINSENSITIVEN,
	SCI_SETMARGINTYPEN, SCI_SETMARGINWIDTHN, SCI_SETMODEVENTMASK,
	SCI_SETMOUSEDOWNCAPTURES, SCI_SETMOUSEDWELLTIME, SCI_SETOVERTYPE,
	SCI_SETPASTECONVERTENDINGS, SCI_SETPRINTCOLOURMODE, SCI_SETPRINTMAGNIFICATION,
	SCI_SETPRINTWRAPMODE, SCI_SETPROPERTY, SCI_SETREADONLY, SCI_SETSAVEPOINT,
	SCI_SETSCROLLWIDTH, SCI_SETSEARCHFLAGS, SCI_SETSEL, SCI_SETSELALPHA,
	SCI_SETSELBACK, SCI_SETSELECTIONEND, SCI_SETSELECTIONMODE,
	SCI_SETSELECTIONSTART, SCI_SETSELFORE, SCI_SETSTATUS, SCI_SETSTYLEBITS,
	SCI_SETSTYLING, SCI_SETSTYLINGEX, SCI_SETTABINDENTS, SCI_SETTABWIDTH,
	SCI_SETTARGETEND, SCI_SETTARGETSTART, SCI_SETTEXT, SCI_SETTWOPHASEDRAW,
	SCI_SETUNDOCOLLECTION, SCI_SETUSEPALETTE, SCI_SETUSETABS, SCI_SETVIEWEOL,
	SCI_SETVIEWWS, SCI_SETVISIBLEPOLICY, SCI_SETVSCROLLBAR, SCI_SETWHITESPACEBACK,
	SCI_SETWHITESPACECHARS, SCI_SETWHITESPACEFORE, SCI_SETWORDCHARS,
	SCI_SETWRAPMODE, SCI_SETWRAPSTARTINDENT, SCI_SETWRAPVISUALFLAGS,
	SCI_SETWRAPVISUALFLAGSLOCATION, SCI_SETXCARETPOLICY, SCI_SETXOFFSET,
	SCI_SETYCARETPOLICY, SCI_SETZOOM, SCI_SHOWLINES, SCI_STARTRECORD,
	SCI_STARTSTYLING, SCI_STOPRECORD, SCI_STYLECLEARALL, SCI_STYLERESETDEFAULT,
	SCI_STYLESETBACK, SCI_STYLESETBOLD, SCI_STYLESETCASE, SCI_STYLESETCHANGEABLE,
	SCI_STYLESETCHARACTERSET, SCI_STYLESETEOLFILLED, SCI_STYLESETFONT,
	SCI_STYLESETFORE, SCI_STYLESETHOTSPOT, SCI_STYLESETITALIC, SCI_STYLESETSIZE,
	SCI_STYLESETUNDERLINE, SCI_STYLESETVISIBLE, SCI_TARGETASUTF8,
	SCI_TARGETFROMSELECTION, SCI_TEXTHEIGHT, SCI_TEXTWIDTH, SCI_TOGGLECARETSTICKY,
	SCI_TOGGLEFOLD, SCI_UNDO, SCI_USEPOPUP, SCI_USERLISTSHOW,
	SCI_VISIBLEFROMDOCLINE, SCI_WORDENDPOSITION, SCI_WORDSTARTPOSITION,
	SCI_WRAPCOUNT, SCI_ZOOMIN, SCI_ZOOMOUT
};

static const int key_cmd[] = {
	// "keyboard" Scintilla commands
	SCI_BACKTAB, SCI_CANCEL, SCI_CHARLEFT, SCI_CHARLEFTEXTEND,
	SCI_CHARLEFTRECTEXTEND, SCI_CHARRIGHT, SCI_CHARRIGHTEXTEND,
	SCI_CHARRIGHTRECTEXTEND, SCI_DELETEBACK, SCI_DELETEBACKNOTLINE,
	SCI_DELLINELEFT, SCI_DELLINERIGHT, SCI_DELWORDLEFT, SCI_DELWORDRIGHT,
	SCI_DOCUMENTEND, SCI_DOCUMENTENDEXTEND, SCI_DOCUMENTSTART,
	SCI_DOCUMENTSTARTEXTEND, SCI_EDITTOGGLEOVERTYPE, SCI_FORMFEED, SCI_HOME,
	SCI_HOMEDISPLAY, SCI_HOMEDISPLAYEXTEND, SCI_HOMEEXTEND, SCI_HOMERECTEXTEND,
	SCI_HOMEWRAP, SCI_HOMEWRAPEXTEND, SCI_LINECOPY, SCI_LINECUT, SCI_LINEDELETE,
	SCI_LINEDOWN, SCI_LINEDOWNEXTEND, SCI_LINEDOWNRECTEXTEND, SCI_LINEDUPLICATE,
	SCI_LINEEND, SCI_LINEENDDISPLAY, SCI_LINEENDDISPLAYEXTEND, SCI_LINEENDEXTEND,
	SCI_LINEENDRECTEXTEND, SCI_LINEENDWRAP, SCI_LINEENDWRAPEXTEND,
	SCI_LINESCROLLDOWN, SCI_LINESCROLLUP, SCI_LINETRANSPOSE, SCI_LINEUP,
	SCI_LINEUPEXTEND, SCI_LINEUPRECTEXTEND, SCI_LOWERCASE, SCI_NEWLINE,
	SCI_PAGEDOWN, SCI_PAGEDOWNEXTEND, SCI_PAGEDOWNRECTEXTEND, SCI_PAGEUP,
	SCI_PAGEUPEXTEND, SCI_PAGEUPRECTEXTEND, SCI_PARADOWN, SCI_PARADOWNEXTEND,
	SCI_PARAUP, SCI_PARAUPEXTEND, SCI_SELECTIONDUPLICATE, SCI_STUTTEREDPAGEDOWN,
	SCI_STUTTEREDPAGEDOWNEXTEND, SCI_STUTTEREDPAGEUP, SCI_STUTTEREDPAGEUPEXTEND,
	SCI_TAB, SCI_UPPERCASE, SCI_VCHOME, SCI_VCHOMEEXTEND, SCI_VCHOMERECTEXTEND,
	SCI_VCHOMEWRAP, SCI_VCHOMEWRAPEXTEND, SCI_WORDLEFT, SCI_WORDLEFTEND,
	SCI_WORDLEFTENDEXTEND, SCI_WORDLEFTEXTEND, SCI_WORDPARTLEFT,
	SCI_WORDPARTLEFTEXTEND, SCI_WORDPARTRIGHT, SCI_WORDPARTRIGHTEXTEND,
	SCI_WORDRIGHT, SCI_WORDRIGHTEND, SCI_WORDRIGHTENDEXTEND, SCI_WORDRIGHTEXTEND
};

// get [GObject] type info
CAMLprim value ml_sci_get_type(value unit)
{
	return Val_GType(scintilla_get_type());
}

#if defined(CAML_DLL) && defined(_WINDOWS_) && defined(REALLY_LATE_BINDING)
typedef value (* vgos_t)(GtkObject* w);
static HMODULE hDllLablGTK2 = NULL;
static FARPROC aVal_GtkObject_sink = NULL;
#endif	/* CAML_DLL && _WINDOWS_ && REALLY_LATE_BINDING */

//	create and return a Scintilla widget
CAMLprim value ml_sci_new(value unit)
{
	CAMLparam0();
	GtkWidget* scintilla_widget = scintilla_new();
#if defined(CAML_DLL) && defined(_WINDOWS_) && defined(REALLY_LATE_BINDING)
	if (hDllLablGTK2 == NULL)
		hDllLablGTK2 = ::LoadLibrary("dlllablgtk2");
	if (hDllLablGTK2 == NULL)
		caml_failwith("ml_sci_new(): dlllablgtk2 can not be loaded");
	if (aVal_GtkObject_sink == NULL)
		aVal_GtkObject_sink = ::GetProcAddress(hDllLablGTK2, "Val_GtkObject_sink");
	if (aVal_GtkObject_sink == NULL)
		caml_failwith("ml_sci_new(): Val_GtkObject_sink not found in dlllablgtk2");
	CAMLreturn(((vgos_t)aVal_GtkObject_sink)(GTK_OBJECT(scintilla_widget)));
#else
	CAMLreturn(Val_GtkWidget_sink(scintilla_widget));
#endif	/* CAML_DLL && _WINDOWS_ && REALLY_LATE_BINDING */
}

//	release platform-specific resources (a mutex in Windows)
CAMLprim value ml_sci_release_resources(value unit)
{
	scintilla_release_resources();
#if defined(CAML_DLL) && defined(_WINDOWS_) && defined(REALLY_LATE_BINDING)
	if (hDllLablGTK2 != NULL)
		::FreeLibrary(hDllLablGTK2), hDllLablGTK2 = NULL, aVal_GtkObject_sink = NULL;
#endif	/* CAML_DLL && _WINDOWS_ && REALLY_LATE_BINDING */
	return Val_unit;
}

/*
	Scintilla message-sending primitives

	The basic model follows Scintilla messaging conventions:

	send_message_xyz obj msg wParam lParam

	Scintilla object, message (an int), "wParam" (almost always an int), and
	"lParam" (usually an int or a string).  To this is added a return type
	specifier (usually an int or unit).  All of this is encoded in the primitive
	name as "_xyz" where x and y are the expected/supported types
	of wParam and lParam, and z is the returned type.

	Possible type-specifiers for x, y,  and z include:
	3	Caml int32
	i	Caml int (31-bit supported now, 63-bit later)
		N.B. - MAY be a bool or sum type... but it is an int to us
	n	Caml nativeint (also basis for "opaque pointers")
	r	Caml record type "text_range" (defined in Scintilla.ml)
	s	Caml string
	u	Caml unit
*/

SS_2(i3i, obj, msg, Int_val, Int32_val, Val_int)
SS_2(i3u, obj, msg, Int_val, Int32_val, Unit)
SS_2(ii3, obj, msg, Int_val, Int_val, Val_int32)
SS_2(iin, obj, msg, Int_val, Int_val, Val_nativeint)
SS_2(iii, obj, msg, Int_val, Int_val, Val_int)
SS_2(iiu, obj, msg, Int_val, Int_val, Unit)
SS_2(inu, obj, msg, Int_val, Nativeint_val, Unit)
SS_2(isi, obj, msg, Int_val, (sptr_t)String_val, Val_int)
SS_2(isu, obj, msg, Int_val, (sptr_t)String_val, Unit)

CAMLprim value ml_sci_send_message_rsi(value obj, value msg, value rVal, value sVal)
{
	CAMLparam4(obj, msg, rVal, sVal);
	TextRange tr = {
		{Int_val(Field(rVal, 0)), Int_val(Field(rVal, 1))}, String_val(sVal)
	};
	CAMLreturn(Val_int(SSM(0, (sptr_t)&tr)));
}

SS_2(sii, obj, msg, (sptr_t)String_val, Int_val, Val_int)
SS_2(ssi, obj, msg, (sptr_t)String_val, (sptr_t)String_val, Val_int)
SS_2(ssu, obj, msg, (sptr_t)String_val, (sptr_t)String_val, Unit)

//	exception case to support find_text (msg == sci_SCI_FINDTEXT)
CAMLprim value ml_sci_send_message_irsr(value obj, value msg, value iVal, value rVal, value sVal)
{
	CAMLparam5(obj, msg, iVal, rVal, sVal);
	CAMLlocal1(tr);
	TextToFind ttf = {
		{Int_val(Field(rVal, 0)), Int_val(Field(rVal, 1))}, String_val(sVal), {0, 0}
	};
	if (SSM(Int_val(iVal), (sptr_t)&ttf) < 0)
		ttf.chrgText.cpMin = -1, ttf.chrgText.cpMax = -1;
	tr = caml_alloc_small(2, 0);
	Field(tr, 0) = Val_int(ttf.chrgText.cpMin), Field(tr, 1) = Val_int(ttf.chrgText.cpMax);
	CAMLreturn(tr);
}

//	exception case to support send_key_cmd ("extended" Scintilla API)
CAMLprim value ml_sci_send_message_key(value obj, value msg)
{
	CAMLparam2(obj, msg);
	scintilla_send_message(ScintillaObject_val(obj), key_cmd[Int_val(msg)], 0, 0);
	CAMLreturn(Val_unit);
}

//	set the Scintilla control ID
CAMLprim value ml_sci_set_id(value obj, value id)
{
	CAMLparam2(obj, id);
	scintilla_set_id(ScintillaObject_val(obj), Int_val(id));
	CAMLreturn(Val_unit);
}

//	return command_msg from an int
// N.B. -  input is as encoded by Scintilla:
//	legal values include 256, 512, 768 in the HIGH 16 bits
CAMLprim value ml_cm_from_int(value cm_idVal)
{
	const int cm = Int_val(cm_idVal) >> 24;
	if (cm == 0 || cm & ~0x03)
		caml_invalid_argument("Scintilla.ml_cm_from_int");
	return Val_int(cm - 1);
}

//	return scnotification from a "Gpointer.boxed option"
//	default to int, use nativeint for "pointers", string for char*
// N.B. - SCNotification.nmhdr.idFrom is now a "ptr" type - we use LOW bits
//	N.B. - MUST use Scintilla >= 1.63 to get "clean" SCNotification struct
CAMLprim value ml_scn_from_boxed_pointer(value bpVal)
{
	CAMLparam1(bpVal);
	CAMLlocal1(ret);
	// (grab a copy of the SCNotification struct)
	SCNotification scn = *(SCNotification*)Field(Field(bpVal, 0), 1);
	ret = caml_alloc_tuple(20);
	const int code = scn.nmhdr.code;
	Field(ret, 0) = caml_copy_nativeint((unsigned long)scn.nmhdr.hwndFrom);
	Field(ret, 1) = Val_int(scn.nmhdr.idFrom);
	Field(ret, 2) = Val_int(code - SCN_STYLENEEDED);
	Field(ret, 3) = Val_int(scn.position);
	Field(ret, 4) = Val_int(scn.ch);
	Field(ret, 5) = Val_emptylist;
	if (scn.modifiers)
		prepend_flags_list(ret, 5, (unsigned int)scn.modifiers);
	Field(ret, 6) = Val_emptylist;
	if (code == SCN_MODIFIED)
		prepend_flags_list(ret, 6, (unsigned int)scn.modificationType);
	if (scn.text != 0) {
		int n;
		// check for special "length in cells" case
		if (code == SCN_MODIFIED
			&& scn.modificationType == (SC_MOD_BEFOREINSERT|SC_PERFORMED_USER))
			n = scn.length * 2;
		// otherwise, use length if present...
		else if (scn.length > 0)
			n = scn.length;
		// ... or just go for zero-termination
		else
			n = strlen(scn.text);
		// do the [string] alloc and copy
		Field(ret, 7) = caml_alloc_string(n),
			memcpy(String_val(Field(ret, 7)), scn.text, n);
	} else
		Field(ret, 7) = caml_copy_string("");
	Field(ret, 8) = Val_int(scn.length);
	Field(ret, 9) = Val_int(scn.linesAdded);
	Field(ret, 10) = Val_int(scn.message);
	Field(ret, 11) = caml_copy_nativeint(scn.wParam);
	Field(ret, 12) = caml_copy_nativeint(scn.lParam);
	Field(ret, 13) = Val_int(scn.line);
	Field(ret, 14) = Val_int(scn.foldLevelNow);
	Field(ret, 15) = Val_int(scn.foldLevelPrev);
	Field(ret, 16) = Val_int(scn.margin);
	Field(ret, 17) = Val_int(scn.listType);
	Field(ret, 18) = Val_int(scn.x);
	Field(ret, 19) = Val_int(scn.y);
	CAMLreturn(ret);
}

//	translate caret_policy(s) (as OCaml sum type list) -> Scintilla caret policy(s)
CAMLprim value ml_int_of_caret_policy_list(value lVal)
{
	static const int t[] = {CARET_SLOP, CARET_STRICT, CARET_JUMPS, CARET_EVEN};
	return Val_int(caml_convert_flag_list(lVal, (int *)t));
}

//	translate char_set (as OCaml sum type) -> Scintilla char set
CAMLprim value ml_int_of_char_set(value iVal)
{
	static const int t[] = {
		SC_CHARSET_8859_15, SC_CHARSET_ANSI, SC_CHARSET_ARABIC, SC_CHARSET_BALTIC,
		SC_CHARSET_CHINESEBIG5, SC_CHARSET_CYRILLIC, SC_CHARSET_DEFAULT,
		SC_CHARSET_EASTEUROPE, SC_CHARSET_GB2312, SC_CHARSET_GREEK,
		SC_CHARSET_HANGUL, SC_CHARSET_HEBREW, SC_CHARSET_JOHAB, SC_CHARSET_MAC,
		SC_CHARSET_OEM, SC_CHARSET_RUSSIAN, SC_CHARSET_SHIFTJIS, SC_CHARSET_SYMBOL,
		SC_CHARSET_THAI, SC_CHARSET_TURKISH, SC_CHARSET_VIETNAMESE
	};
	return Val_int(t[Int_val(iVal)]);
}

//	translate key_cmd (as OCaml sum type) -> Scintilla "keyboard" command
CAMLprim value ml_int_of_key_cmd(value iVal)
{
	return Val_int(key_cmd[Int_val(iVal)]);
}

//	translate key_code (as OCaml sum type) -> Scintilla key code
CAMLprim value ml_int_of_key_code(value iVal)
{
	static const int t[] = {
		SCK_ADD, SCK_BACK, SCK_DELETE, SCK_DIVIDE, SCK_DOWN, SCK_END, SCK_ESCAPE,
		SCK_HOME, SCK_INSERT, SCK_LEFT, SCK_MENU, SCK_NEXT, SCK_PRIOR, SCK_RETURN,
		SCK_RIGHT, SCK_RWIN, SCK_SUBTRACT, SCK_TAB, SCK_UP, SCK_WIN
	};
	return Val_int(t[Int_val(iVal)]);
}

//	translate key_modifier(s) (as OCaml sum type list) -> Scintilla key modifier(s)
CAMLprim value ml_int_of_key_modifier_list(value lVal)
{
	static const int t[] = {
		/* SCMOD_NORM, */ SCMOD_ALT, SCMOD_CTRL, SCMOD_SHIFT
	};
	return Val_int(caml_convert_flag_list(lVal, (int *)t));
}

//	translate marker_symbol (as OCaml sum type) -> Scintilla marker symbol
CAMLprim value ml_int_of_marker_symbol(value mVal)
{
	CAMLparam1(mVal);
	CAMLlocal1(ret);
	ret = Is_block(mVal)?
		Val_int(Int_val(Field(mVal, 0)) + SC_MARK_CHARACTER): mVal;
	CAMLreturn(ret);
}

//	translate modification_type(s) (as OCaml sum type list) -> Scintilla modification_type(s)
CAMLprim value ml_int_of_modification_type_list(value lVal)
{
	static const int t[] = {
		SC_MOD_INSERTTEXT, SC_MOD_DELETETEXT, SC_MOD_CHANGESTYLE,
		SC_MOD_CHANGEFOLD, SC_PERFORMED_USER, SC_PERFORMED_UNDO,
		SC_PERFORMED_REDO, SC_MULTISTEPUNDOREDO, SC_LASTSTEPINUNDOREDO,
		SC_MOD_CHANGEMARKER, SC_MOD_BEFOREINSERT, SC_MOD_BEFOREDELETE,
		SC_MULTILINEUNDOREDO, SC_STARTACTION /*, SC_MODEVENTMASKALL */
	};
	return Val_int(caml_convert_flag_list(lVal, (int *)t));
}

static const int search_flag[] = {
	SCFIND_MATCHCASE, SCFIND_WHOLEWORD, SCFIND_WORDSTART,
	SCFIND_REGEXP, SCFIND_POSIX
};

//	translate search_flag(s) (as OCaml sum type list) -> Scintilla search flag(s)
CAMLprim value ml_int_of_search_flag_list(value lVal)
{
	return Val_int(caml_convert_flag_list(lVal, (int *)search_flag));
}

//	translate Scintilla search flag -> search_flag (as OCaml sum type)
CAMLprim value ml_search_flag_of_int(value iVal)
{
	const int e = Unsigned_int_val(iVal);
	for (int i = 0; i < 5; i++)
		if (search_flag[i] == e)
			return Val_int(i);
	caml_invalid_argument("Scintilla.ml_search_flag_of_int");
	return Val_unit;	// ("dummy" since not all compilers "get" Noreturn)
}

//	translate visible_policy(s) (as OCaml sum type list) -> Scintilla visible policy(s)
CAMLprim value ml_int_of_visible_policy_list(value lVal)
{
	static const int t[] = {VISIBLE_SLOP, VISIBLE_STRICT};
	return Val_int(caml_convert_flag_list(lVal, (int *)t));
}

//	translate visual_flag(s) (as OCaml sum type list) -> Scintilla visual_flag(s)
CAMLprim value ml_int_of_visual_flag_list(value lVal)
{
	static const int t[] = {
		SC_WRAPVISUALFLAG_NONE, SC_WRAPVISUALFLAG_END, SC_WRAPVISUALFLAG_START
	};
	return Val_int(caml_convert_flag_list(lVal, (int *)t));
}

//	translate visual_flag_loc(s) (as OCaml sum type list) -> Scintilla visual_flag_loc(s)
CAMLprim value ml_int_of_visual_flag_loc_list(value lVal)
{
	static const int t[] = {
		SC_WRAPVISUALFLAGLOC_DEFAULT, SC_WRAPVISUALFLAGLOC_END_BY_TEXT,
		SC_WRAPVISUALFLAGLOC_START_BY_TEXT
	};
	return Val_int(caml_convert_flag_list(lVal, (int *)t));
}

// prepend [any] flags from supplied mask to list in specified aggregate slot
// N.B. - this works because flag vals exactly parallel sum type declarations
static void prepend_flags_list(value& agg, int slot, unsigned int m)
{
	for (int e = 0; m; e++, m >>= 1)
		if (m & 1) {
			const value cons = caml_alloc_small(2, Tag_cons);
			Field(cons, 0) = Val_int(e), Field(cons, 1) = Field(agg, slot),
				Field(agg, slot) = cons;
		}
}

}
