(*
	LablScintilla - Scintilla [LablGTK] binding for Objective Caml

	author: Robert Roessler
	email: robertr@rftp.com
	web: http://www.rftp.com

--------------------------------------------------------------------------------
Copyright (c) 2005-2007 Robert Roessler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Robert Roessler nor the names of its contributors
      may be used to endorse or promote products derived from this
      software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
--------------------------------------------------------------------------------
*)

(* scintilla.mli - interface of Scintilla [LablGTK] binding for OCaml *)

type scintilla = [ `container | `gtk | `scintilla | `widget ]

(* mapping of "core" Scintilla types *)
type opaque_ptr
type document
type text_range = { cpMin : int; cpMax : int; }

type cache_mode =
	| SC_CACHE_NONE | SC_CACHE_CARET | SC_CACHE_PAGE | SC_CACHE_DOCUMENT

type caret_policy = CARET_SLOP | CARET_STRICT | CARET_JUMPS | CARET_EVEN

type case_mode = SC_CASE_MIXED | SC_CASE_UPPER | SC_CASE_LOWER

type char_set =
	| SC_CHARSET_8859_15 | SC_CHARSET_ANSI | SC_CHARSET_ARABIC | SC_CHARSET_BALTIC
	| SC_CHARSET_CHINESEBIG5 | SC_CHARSET_CYRILLIC | SC_CHARSET_DEFAULT
	| SC_CHARSET_EASTEUROPE | SC_CHARSET_GB2312 | SC_CHARSET_GREEK
	| SC_CHARSET_HANGUL | SC_CHARSET_HEBREW | SC_CHARSET_JOHAB | SC_CHARSET_MAC
	| SC_CHARSET_OEM | SC_CHARSET_RUSSIAN | SC_CHARSET_SHIFTJIS | SC_CHARSET_SYMBOL
	| SC_CHARSET_THAI | SC_CHARSET_TURKISH | SC_CHARSET_VIETNAMESE

type command_msg = SCEN_KILLFOCUS | SCEN_SETFOCUS | SCEN_CHANGE

type edge_mode = EDGE_NONE | EDGE_LINE | EDGE_BACKGROUND

type eol_mode = SC_EOL_CR | SC_EOL_CRLF | SC_EOL_LF

type indicator_style = 
	| INDIC_PLAIN | INDIC_SQUIGGLE | INDIC_TT | INDIC_DIAGONAL
	| INDIC_STRIKE | INDIC_HIDDEN | INDIC_BOX | INDIC_ROUNDBOX

type key_code =
	| SCK_ADD | SCK_BACK | SCK_DELETE | SCK_DIVIDE | SCK_DOWN | SCK_END
	| SCK_ESCAPE | SCK_HOME | SCK_INSERT | SCK_LEFT | SCK_MENU | SCK_NEXT
	| SCK_PRIOR | SCK_RETURN | SCK_RIGHT | SCK_RWIN | SCK_SUBTRACT | SCK_TAB
	| SCK_UP | SCK_WIN

type key_modifier = (* SCMOD_NORM | *) SCMOD_SHIFT | SCMOD_CTRL | SCMOD_ALT

type key_definition = KeyDef of key_code * key_modifier list

type margin_type = SC_MARGIN_SYMBOL | SC_MARGIN_NUMBER | SC_MARGIN_BACK
	| SC_MARGIN_FORE

type marker_symbol =
	| SC_MARK_CIRCLE | SC_MARK_ROUNDRECT | SC_MARK_ARROW | SC_MARK_SMALLRECT
	| SC_MARK_SHORTARROW | SC_MARK_EMPTY | SC_MARK_ARROWDOWN | SC_MARK_MINUS
	| SC_MARK_PLUS | SC_MARK_VLINE | SC_MARK_LCORNER | SC_MARK_TCORNER
	| SC_MARK_BOXPLUS | SC_MARK_BOXPLUSCONNECTED | SC_MARK_BOXMINUS
	| SC_MARK_BOXMINUSCONNECTED | SC_MARK_LCORNERCURVE | SC_MARK_TCORNERCURVE
	| SC_MARK_CIRCLEPLUS | SC_MARK_CIRCLEPLUSCONNECTED | SC_MARK_CIRCLEMINUS
	| SC_MARK_CIRCLEMINUSCONNECTED | SC_MARK_BACKGROUND | SC_MARK_DOTDOTDOT
	| SC_MARK_ARROWS | SC_MARK_PIXMAP | SC_MARK_FULLRECT
	| SC_MARK_CHARACTER of char

type modification_type =
	| SC_MOD_INSERTTEXT | SC_MOD_DELETETEXT | SC_MOD_CHANGESTYLE
	| SC_MOD_CHANGEFOLD | SC_PERFORMED_USER | SC_PERFORMED_UNDO
	| SC_PERFORMED_REDO | SC_MULTISTEPUNDOREDO | SC_LASTSTEPINUNDOREDO
	| SC_MOD_CHANGEMARKER | SC_MOD_BEFOREINSERT | SC_MOD_BEFOREDELETE
	| SC_MULTILINEUNDOREDO | SC_STARTACTION (* | SC_MODEVENTMASKALL *)

(* N.B. - leave in NUMERIC order (and SCN_POSCHANGED is a "place-holder") *)
type notification_msg =
	| SCN_STYLENEEDED | SCN_CHARADDED | SCN_SAVEPOINTREACHED | SCN_SAVEPOINTLEFT
	| SCN_MODIFYATTEMPTRO | SCN_KEY | SCN_DOUBLECLICK | SCN_UPDATEUI
	| SCN_MODIFIED | SCN_MACRORECORD | SCN_MARGINCLICK | SCN_NEEDSHOWN
	| SCN_POSCHANGED | SCN_PAINTED | SCN_USERLISTSELECTION | SCN_URIDROPPED
	| SCN_DWELLSTART | SCN_DWELLEND | SCN_ZOOM | SCN_HOTSPOTCLICK
	| SCN_HOTSPOTDOUBLECLICK | SCN_CALLTIPCLICK | SCN_AUTOCSELECTION 

type search_flag =
	| SCFIND_MATCHCASE | SCFIND_WHOLEWORD | SCFIND_WORDSTART
	| SCFIND_REGEXP | SCFIND_POSIX

type selection_mode = SC_SEL_STREAM | SC_SEL_RECTANGLE | SC_SEL_LINES

type visible_policy = VISIBLE_SLOP | VISIBLE_STRICT

type visual_flag =
	| SC_WRAPVISUALFLAG_NONE | SC_WRAPVISUALFLAG_END | SC_WRAPVISUALFLAG_START

type visual_flag_loc =
	| SC_WRAPVISUALFLAGLOC_DEFAULT | SC_WRAPVISUALFLAGLOC_END_BY_TEXT
	| SC_WRAPVISUALFLAGLOC_START_BY_TEXT

type white_space_mode =
	| SCWS_INVISIBLE | SCWS_VISIBLEALWAYS | SCWS_VISIBLEAFTERINDENT

type wrap_mode = SC_WRAP_NONE | SC_WRAP_WORD | SC_WRAP_CHAR

(* "standard" Scintilla commands *)
type sci_cmd =
	| SCI_ADDREFDOCUMENT | SCI_ADDSTYLEDTEXT | SCI_ADDTEXT | SCI_ALLOCATE
	| SCI_APPENDTEXT | SCI_ASSIGNCMDKEY | SCI_AUTOCACTIVE | SCI_AUTOCCANCEL
	| SCI_AUTOCCOMPLETE | SCI_AUTOCGETAUTOHIDE | SCI_AUTOCGETCANCELATSTART
	| SCI_AUTOCGETCHOOSESINGLE | SCI_AUTOCGETCURRENT | SCI_AUTOCGETDROPRESTOFWORD
	| SCI_AUTOCGETIGNORECASE | SCI_AUTOCGETMAXHEIGHT | SCI_AUTOCGETMAXWIDTH
	| SCI_AUTOCGETSEPARATOR | SCI_AUTOCGETTYPESEPARATOR | SCI_AUTOCPOSSTART
	| SCI_AUTOCSELECT | SCI_AUTOCSETAUTOHIDE | SCI_AUTOCSETCANCELATSTART
	| SCI_AUTOCSETCHOOSESINGLE | SCI_AUTOCSETDROPRESTOFWORD | SCI_AUTOCSETFILLUPS
	| SCI_AUTOCSETIGNORECASE | SCI_AUTOCSETMAXHEIGHT | SCI_AUTOCSETMAXWIDTH
	| SCI_AUTOCSETSEPARATOR | SCI_AUTOCSETTYPESEPARATOR | SCI_AUTOCSHOW
	| SCI_AUTOCSTOPS | SCI_BEGINUNDOACTION | SCI_BRACEBADLIGHT
	| SCI_BRACEHIGHLIGHT | SCI_BRACEMATCH | SCI_CALLTIPACTIVE | SCI_CALLTIPCANCEL
	| SCI_CALLTIPPOSSTART | SCI_CALLTIPSETBACK | SCI_CALLTIPSETFORE
	| SCI_CALLTIPSETFOREHLT | SCI_CALLTIPSETHLT | SCI_CALLTIPSHOW
	| SCI_CALLTIPUSESTYLE | SCI_CANPASTE | SCI_CANREDO | SCI_CANUNDO
	| SCI_CHOOSECARETX | SCI_CLEAR | SCI_CLEARALL | SCI_CLEARALLCMDKEYS
	| SCI_CLEARCMDKEY | SCI_CLEARDOCUMENTSTYLE | SCI_CLEARREGISTEREDIMAGES
	| SCI_COLOURISE | SCI_CONVERTEOLS | SCI_COPY | SCI_COPYRANGE | SCI_COPYTEXT
	| SCI_CREATEDOCUMENT | SCI_CUT | SCI_DOCLINEFROMVISIBLE | SCI_EMPTYUNDOBUFFER
	| SCI_ENCODEDFROMUTF8 | SCI_ENDUNDOACTION | SCI_ENSUREVISIBLE
	| SCI_ENSUREVISIBLEENFORCEPOLICY | SCI_FINDCOLUMN | SCI_FINDTEXT
	| SCI_FORMATRANGE | SCI_GETANCHOR | SCI_GETBACKSPACEUNINDENTS
	| SCI_GETBUFFEREDDRAW | SCI_GETCARETFORE | SCI_GETCARETLINEBACK
	| SCI_GETCARETLINEBACKALPHA | SCI_GETCARETLINEVISIBLE | SCI_GETCARETPERIOD
	| SCI_GETCARETSTICKY | SCI_GETCARETWIDTH | SCI_GETCHARAT | SCI_GETCODEPAGE
	| SCI_GETCOLUMN | SCI_GETCONTROLCHARSYMBOL | SCI_GETCURLINE
	| SCI_GETCURRENTPOS | SCI_GETCURSOR | SCI_GETDIRECTFUNCTION
	| SCI_GETDIRECTPOINTER | SCI_GETDOCPOINTER | SCI_GETEDGECOLOUR
	| SCI_GETEDGECOLUMN | SCI_GETEDGEMODE | SCI_GETENDATLASTLINE
	| SCI_GETENDSTYLED | SCI_GETEOLMODE | SCI_GETFIRSTVISIBLELINE | SCI_GETFOCUS
	| SCI_GETFOLDEXPANDED | SCI_GETFOLDLEVEL | SCI_GETFOLDPARENT
	| SCI_GETHIGHLIGHTGUIDE | SCI_GETHSCROLLBAR | SCI_GETINDENT
	| SCI_GETINDENTATIONGUIDES | SCI_GETLASTCHILD | SCI_GETLAYOUTCACHE
	| SCI_GETLENGTH | SCI_GETLEXER | SCI_GETLINE | SCI_GETLINECOUNT
	| SCI_GETLINEENDPOSITION | SCI_GETLINEINDENTATION | SCI_GETLINEINDENTPOSITION
	| SCI_GETLINESELENDPOSITION | SCI_GETLINESELSTARTPOSITION | SCI_GETLINESTATE
	| SCI_GETLINEVISIBLE | SCI_GETMARGINLEFT | SCI_GETMARGINMASKN
	| SCI_GETMARGINRIGHT | SCI_GETMARGINSENSITIVEN | SCI_GETMARGINTYPEN
	| SCI_GETMARGINWIDTHN | SCI_GETMAXLINESTATE | SCI_GETMODEVENTMASK
	| SCI_GETMODIFY | SCI_GETMOUSEDOWNCAPTURES | SCI_GETMOUSEDWELLTIME
	| SCI_GETOVERTYPE | SCI_GETPASTECONVERTENDINGS | SCI_GETPRINTCOLOURMODE
	| SCI_GETPRINTMAGNIFICATION | SCI_GETPRINTWRAPMODE | SCI_GETPROPERTY
	| SCI_GETPROPERTYEXPANDED | SCI_GETPROPERTYINT | SCI_GETREADONLY
	| SCI_GETSCROLLWIDTH | SCI_GETSEARCHFLAGS | SCI_GETSELALPHA
	| SCI_GETSELECTIONEND | SCI_GETSELECTIONMODE | SCI_GETSELECTIONSTART
	| SCI_GETSELTEXT | SCI_GETSTATUS | SCI_GETSTYLEAT | SCI_GETSTYLEBITS
	| SCI_GETSTYLEBITSNEEDED | SCI_GETSTYLEDTEXT | SCI_GETTABINDENTS
	| SCI_GETTABWIDTH | SCI_GETTARGETEND | SCI_GETTARGETSTART | SCI_GETTEXT
	| SCI_GETTEXTLENGTH | SCI_GETTEXTRANGE | SCI_GETTWOPHASEDRAW
	| SCI_GETUNDOCOLLECTION | SCI_GETUSEPALETTE | SCI_GETUSETABS | SCI_GETVIEWEOL
	| SCI_GETVIEWWS | SCI_GETVSCROLLBAR | SCI_GETWRAPMODE | SCI_GETWRAPSTARTINDENT
	| SCI_GETWRAPVISUALFLAGS | SCI_GETWRAPVISUALFLAGSLOCATION | SCI_GETXOFFSET
	| SCI_GETZOOM | SCI_GOTOLINE | SCI_GOTOPOS | SCI_GRABFOCUS | SCI_HIDELINES
	| SCI_HIDESELECTION | SCI_INDICGETFORE | SCI_INDICGETSTYLE | SCI_INDICSETFORE
	| SCI_INDICSETSTYLE | SCI_INSERTTEXT | SCI_LINEFROMPOSITION | SCI_LINELENGTH
	| SCI_LINESCROLL | SCI_LINESJOIN | SCI_LINESONSCREEN | SCI_LINESSPLIT
	| SCI_LOADLEXERLIBRARY | SCI_MARKERADD | SCI_MARKERADDSET | SCI_MARKERDEFINE
	| SCI_MARKERDEFINEPIXMAP | SCI_MARKERDELETE | SCI_MARKERDELETEALL
	| SCI_MARKERDELETEHANDLE | SCI_MARKERGET | SCI_MARKERLINEFROMHANDLE
	| SCI_MARKERNEXT | SCI_MARKERPREVIOUS | SCI_MARKERSETALPHA | SCI_MARKERSETBACK
	| SCI_MARKERSETFORE | SCI_MOVECARETINSIDEVIEW | SCI_NULL | SCI_PASTE
	| SCI_POINTXFROMPOSITION | SCI_POINTYFROMPOSITION | SCI_POSITIONAFTER
	| SCI_POSITIONBEFORE | SCI_POSITIONFROMLINE | SCI_POSITIONFROMPOINT
	| SCI_POSITIONFROMPOINTCLOSE | SCI_REDO | SCI_REGISTERIMAGE
	| SCI_RELEASEDOCUMENT | SCI_REPLACESEL | SCI_REPLACETARGET
	| SCI_REPLACETARGETRE | SCI_SCROLLCARET | SCI_SEARCHANCHOR
	| SCI_SEARCHINTARGET | SCI_SEARCHNEXT | SCI_SEARCHPREV | SCI_SELECTALL
	| SCI_SELECTIONISRECTANGLE | SCI_SETANCHOR | SCI_SETBACKSPACEUNINDENTS
	| SCI_SETBUFFEREDDRAW | SCI_SETCARETFORE | SCI_SETCARETLINEBACK
	| SCI_SETCARETLINEBACKALPHA | SCI_SETCARETLINEVISIBLE | SCI_SETCARETPERIOD
	| SCI_SETCARETSTICKY | SCI_SETCARETWIDTH | SCI_SETCHARSDEFAULT
	| SCI_SETCODEPAGE | SCI_SETCONTROLCHARSYMBOL | SCI_SETCURRENTPOS
	| SCI_SETCURSOR | SCI_SETDOCPOINTER | SCI_SETEDGECOLOUR | SCI_SETEDGECOLUMN
	| SCI_SETEDGEMODE | SCI_SETENDATLASTLINE | SCI_SETEOLMODE | SCI_SETFOCUS
	| SCI_SETFOLDEXPANDED | SCI_SETFOLDFLAGS | SCI_SETFOLDLEVEL
	| SCI_SETFOLDMARGINCOLOUR | SCI_SETFOLDMARGINHICOLOUR | SCI_SETHIGHLIGHTGUIDE
	| SCI_SETHOTSPOTACTIVEBACK | SCI_SETHOTSPOTACTIVEFORE
	| SCI_SETHOTSPOTACTIVEUNDERLINE | SCI_SETHOTSPOTSINGLELINE | SCI_SETHSCROLLBAR
	| SCI_SETINDENT | SCI_SETINDENTATIONGUIDES | SCI_SETKEYWORDS
	| SCI_SETLAYOUTCACHE | SCI_SETLENGTHFORENCODE | SCI_SETLEXER
	| SCI_SETLEXERLANGUAGE | SCI_SETLINEINDENTATION | SCI_SETLINESTATE
	| SCI_SETMARGINLEFT | SCI_SETMARGINMASKN | SCI_SETMARGINRIGHT
	| SCI_SETMARGINSENSITIVEN | SCI_SETMARGINTYPEN | SCI_SETMARGINWIDTHN
	| SCI_SETMODEVENTMASK | SCI_SETMOUSEDOWNCAPTURES | SCI_SETMOUSEDWELLTIME
	| SCI_SETOVERTYPE | SCI_SETPASTECONVERTENDINGS | SCI_SETPRINTCOLOURMODE
	| SCI_SETPRINTMAGNIFICATION | SCI_SETPRINTWRAPMODE | SCI_SETPROPERTY
	| SCI_SETREADONLY | SCI_SETSAVEPOINT | SCI_SETSCROLLWIDTH | SCI_SETSEARCHFLAGS
	| SCI_SETSEL | SCI_SETSELALPHA | SCI_SETSELBACK | SCI_SETSELECTIONEND
	| SCI_SETSELECTIONMODE | SCI_SETSELECTIONSTART | SCI_SETSELFORE
	| SCI_SETSTATUS | SCI_SETSTYLEBITS | SCI_SETSTYLING | SCI_SETSTYLINGEX
	| SCI_SETTABINDENTS | SCI_SETTABWIDTH | SCI_SETTARGETEND | SCI_SETTARGETSTART
	| SCI_SETTEXT | SCI_SETTWOPHASEDRAW | SCI_SETUNDOCOLLECTION
	| SCI_SETUSEPALETTE | SCI_SETUSETABS | SCI_SETVIEWEOL | SCI_SETVIEWWS
	| SCI_SETVISIBLEPOLICY | SCI_SETVSCROLLBAR | SCI_SETWHITESPACEBACK
	| SCI_SETWHITESPACECHARS | SCI_SETWHITESPACEFORE | SCI_SETWORDCHARS
	| SCI_SETWRAPMODE | SCI_SETWRAPSTARTINDENT | SCI_SETWRAPVISUALFLAGS
	| SCI_SETWRAPVISUALFLAGSLOCATION | SCI_SETXCARETPOLICY | SCI_SETXOFFSET
	| SCI_SETYCARETPOLICY | SCI_SETZOOM | SCI_SHOWLINES | SCI_STARTRECORD
	| SCI_STARTSTYLING | SCI_STOPRECORD | SCI_STYLECLEARALL
	| SCI_STYLERESETDEFAULT | SCI_STYLESETBACK | SCI_STYLESETBOLD
	| SCI_STYLESETCASE | SCI_STYLESETCHANGEABLE | SCI_STYLESETCHARACTERSET
	| SCI_STYLESETEOLFILLED | SCI_STYLESETFONT | SCI_STYLESETFORE
	| SCI_STYLESETHOTSPOT | SCI_STYLESETITALIC | SCI_STYLESETSIZE
	| SCI_STYLESETUNDERLINE | SCI_STYLESETVISIBLE | SCI_TARGETASUTF8
	| SCI_TARGETFROMSELECTION | SCI_TEXTHEIGHT | SCI_TEXTWIDTH
	| SCI_TOGGLECARETSTICKY | SCI_TOGGLEFOLD | SCI_UNDO | SCI_USEPOPUP
	| SCI_USERLISTSHOW | SCI_VISIBLEFROMDOCLINE | SCI_WORDENDPOSITION
	| SCI_WORDSTARTPOSITION | SCI_WRAPCOUNT | SCI_ZOOMIN | SCI_ZOOMOUT

(* "keyboard" Scintilla commands *)
type key_cmd =
	| SCI_BACKTAB | SCI_CANCEL | SCI_CHARLEFT | SCI_CHARLEFTEXTEND
	| SCI_CHARLEFTRECTEXTEND | SCI_CHARRIGHT | SCI_CHARRIGHTEXTEND
	| SCI_CHARRIGHTRECTEXTEND | SCI_DELETEBACK | SCI_DELETEBACKNOTLINE
	| SCI_DELLINELEFT | SCI_DELLINERIGHT | SCI_DELWORDLEFT | SCI_DELWORDRIGHT
	| SCI_DOCUMENTEND | SCI_DOCUMENTENDEXTEND | SCI_DOCUMENTSTART
	| SCI_DOCUMENTSTARTEXTEND | SCI_EDITTOGGLEOVERTYPE | SCI_FORMFEED | SCI_HOME
	| SCI_HOMEDISPLAY | SCI_HOMEDISPLAYEXTEND | SCI_HOMEEXTEND
	| SCI_HOMERECTEXTEND | SCI_HOMEWRAP | SCI_HOMEWRAPEXTEND | SCI_LINECOPY
	| SCI_LINECUT | SCI_LINEDELETE | SCI_LINEDOWN | SCI_LINEDOWNEXTEND
	| SCI_LINEDOWNRECTEXTEND | SCI_LINEDUPLICATE | SCI_LINEEND
	| SCI_LINEENDDISPLAY | SCI_LINEENDDISPLAYEXTEND | SCI_LINEENDEXTEND
	| SCI_LINEENDRECTEXTEND | SCI_LINEENDWRAP | SCI_LINEENDWRAPEXTEND
	| SCI_LINESCROLLDOWN | SCI_LINESCROLLUP | SCI_LINETRANSPOSE | SCI_LINEUP
	| SCI_LINEUPEXTEND | SCI_LINEUPRECTEXTEND | SCI_LOWERCASE | SCI_NEWLINE
	| SCI_PAGEDOWN | SCI_PAGEDOWNEXTEND | SCI_PAGEDOWNRECTEXTEND | SCI_PAGEUP
	| SCI_PAGEUPEXTEND | SCI_PAGEUPRECTEXTEND | SCI_PARADOWN | SCI_PARADOWNEXTEND
	| SCI_PARAUP | SCI_PARAUPEXTEND | SCI_SELECTIONDUPLICATE
	| SCI_STUTTEREDPAGEDOWN | SCI_STUTTEREDPAGEDOWNEXTEND | SCI_STUTTEREDPAGEUP
	| SCI_STUTTEREDPAGEUPEXTEND | SCI_TAB | SCI_UPPERCASE | SCI_VCHOME
	| SCI_VCHOMEEXTEND | SCI_VCHOMERECTEXTEND | SCI_VCHOMEWRAP
	| SCI_VCHOMEWRAPEXTEND | SCI_WORDLEFT | SCI_WORDLEFTEND
	| SCI_WORDLEFTENDEXTEND | SCI_WORDLEFTEXTEND | SCI_WORDPARTLEFT
	| SCI_WORDPARTLEFTEXTEND | SCI_WORDPARTRIGHT | SCI_WORDPARTRIGHTEXTEND
	| SCI_WORDRIGHT | SCI_WORDRIGHTEND | SCI_WORDRIGHTENDEXTEND
	| SCI_WORDRIGHTEXTEND

(* N.B. - these are ONLY created by ml_scn_from_boxed_pointer *)
type scnotification = private {
	hwndFrom: opaque_ptr; idFrom: int; code: notification_msg;
	position: int;
	ch: int;
	modifiers: key_modifier list;
	modificationType: modification_type list;
	text: string; length: int;
	linesAdded: int;
	message: int;
	wParam: opaque_ptr; lParam: opaque_ptr;
	line: int;
	foldLevelNow: int; foldLevelPrev: int;
	margin: int;
	x: int; y: int
}

(* mapping of NON-typed/integer Scintilla constants *)
val sci_SC_CP_DBCS : int
val sci_SC_CP_UTF8 : int

val sci_SC_CURSORNORMAL : int
val sci_SC_CURSORWAIT : int

val sci_SC_TIME_FOREVER : int

val sci_INVALID_POSITION : int

val sci_SC_MARKNUM_FOLDEREND : int
val sci_SC_MARKNUM_FOLDEROPENMID : int
val sci_SC_MARKNUM_FOLDERMIDTAIL : int
val sci_SC_MARKNUM_FOLDERTAIL : int
val sci_SC_MARKNUM_FOLDERSUB : int
val sci_SC_MARKNUM_FOLDER : int
val sci_SC_MARKNUM_FOLDEROPEN : int

val sci_SC_MASK_FOLDERS : int32

val sci_SC_FOLDLEVELBASE : int
val sci_SC_FOLDLEVELWHITEFLAG : int
val sci_SC_FOLDLEVELHEADERFLAG : int
val sci_SC_FOLDLEVELBOXHEADERFLAG : int
val sci_SC_FOLDLEVELBOXFOOTERFLAG : int
val sci_SC_FOLDLEVELCONTRACTED : int
val sci_SC_FOLDLEVELUNINDENT : int
val sci_SC_FOLDLEVELNUMBERMASK : int

val sci_STYLE_DEFAULT : int
val sci_STYLE_LINENUMBER : int
val sci_STYLE_BRACELIGHT : int
val sci_STYLE_BRACEBAD : int
val sci_STYLE_CONTROLCHAR : int
val sci_STYLE_INDENTGUIDE : int
val sci_STYLE_CALLTIP : int
val sci_STYLE_LASTPREDEFINED : int
val sci_STYLE_MAX : int

val sci_INDIC0_MASK : int
val sci_INDIC1_MASK : int
val sci_INDIC2_MASK : int
val sci_INDICS_MASK : int

val sci_SC_ALPHA_TRANSPARENT : int
val sci_SC_ALPHA_OPAQUE : int
val sci_SC_ALPHA_NOALPHA : int

(* define the "signals" interface for the Scintilla widget *)
(* N.B. - nearly everything of interest will appear as "sci-notify" signals *)
module S :
	sig
		(* callback: int -> command_msg -> Gpointer.boxed option -> unit *)
		val command :
		([> `scintilla ], int -> command_msg -> Gpointer.boxed option -> unit)
			GtkSignal.t
		(* callback: int -> scnotification -> unit *)
		val notify : ([> `scintilla ], int -> scnotification -> unit) GtkSignal.t
	end

class virtual scintilla_sigs :
	object
		method private virtual connect :
			'a. ([> `scintilla ], 'a) GtkSignal.t -> callback:'a -> GtkSignal.id

		method command :
			callback:(int -> command_msg -> Gpointer.boxed option -> unit) ->
			GtkSignal.id
		method notify : callback:(int -> scnotification -> unit) -> GtkSignal.id
	end

class scintilla_signals :
	([> scintilla ] as 'b) Gtk.obj ->
	object ('a)
		method after : 'a
		method destroy : callback:(unit -> unit) -> GtkSignal.id

		method command :
			callback:(int -> command_msg -> Gpointer.boxed option -> unit) ->
			GtkSignal.id
		method notify : callback:(int -> scnotification -> unit) -> GtkSignal.id
	end

(* the Scintilla editor widget wrapper class *)
class editor :
	scintilla Gtk.obj ->
	object
		method as_widget : Gtk.widget Gtk.obj
		method coerce : GObj.widget
		method destroy : unit -> unit
		method drag : GObj.drag_ops
		method get_oid : int
		method misc : GObj.misc_ops
		method connect : scintilla_signals

		method add_ref_document : document -> unit
		method add_styled_text : int -> string -> unit
		method add_text : int -> string -> unit
		method allocate : int -> unit
		method append_text : int -> string -> unit
		method assign_cmd_key : key_definition -> key_cmd -> unit
		method autoc_active : int
		method autoc_cancel : unit -> unit
		method autoc_complete : unit -> unit
		method autoc_get_auto_hide : bool
		method autoc_get_cancel_at_start : bool
		method autoc_get_choose_single : bool
		method autoc_get_current : int
		method autoc_get_drop_rest_of_word : bool
		method autoc_get_ignore_case : bool
		method autoc_get_separator : char
		method autoc_get_type_separator : char
		method autoc_pos_start : int
		method autoc_select : int -> unit
		method autoc_set_auto_hide : bool -> unit
		method autoc_set_cancel_at_start : bool -> unit
		method autoc_set_choose_single : bool -> unit
		method autoc_set_drop_rest_of_word : bool -> unit
		method autoc_set_fillups : int -> unit
		method autoc_set_ignore_case : bool -> unit
		method autoc_set_separator : int -> unit
		method autoc_set_type_separator : int -> unit
		method autoc_show : int -> int -> unit
		method autoc_stop_s : int -> unit
		method begin_undo_action : unit -> unit
		method brace_badlight : int -> unit
		method brace_highlight : int -> int -> unit
		method brace_match : int -> int -> int
		method call_tip_active : int
		method call_tip_cancel : unit -> unit
		method call_tip_pos_start : int
		method call_tip_set_back : int -> unit
		method call_tip_set_fore : int -> unit
		method call_tip_set_fore_hlt : int -> unit
		method call_tip_set_hlt : int -> int -> unit
		method call_tip_show : int -> string -> unit
		method can_paste : int
		method can_redo : int
		method can_undo : int
		method choose_caret_x : unit -> unit
		method clear : unit -> unit
		method clear_all : unit -> unit
		method clear_all_cmd_keys : unit -> unit
		method clear_cmd_key : key_definition -> unit
		method clear_document_style : unit -> unit
		method clear_registered_images : unit -> unit
		method colourise : int -> int -> unit
		method convert_eols : eol_mode -> unit
		method copy : unit -> unit
		method copy_range : int -> int -> unit
		method copy_text : int -> string -> unit
		method create_document : unit -> document
		method cut : unit -> unit
		method doc_line_from_visible : int -> int
		method empty_undo_buffer : unit -> unit
		method encoded_from_utf8 : string -> string
		method end_undo_action : unit -> unit
		method ensure_visible : int -> unit
		method ensure_visible_enforce_policy : int -> unit
		method find_column : int -> int -> int
		method find_text :
			?flags:search_flag list -> text_range -> string -> text_range
		method get_anchor : int
		method get_backspace_unindents : bool
		method get_buffered_draw : bool
		method get_caret_fore : int
		method get_caret_line_back : int
		method get_caret_line_back_alpha : int
		method get_caret_line_visible : bool
		method get_caret_period : int
		method get_caret_width : int
		method get_char_at : int -> char
		method get_code_page : int
		method get_column : int -> int
		method get_control_char_symbol : char
		method get_cur_line : string
		method get_current_pos : int
		method get_cursor : int
		method get_doc_pointer : document
		method get_edge_colour : int
		method get_edge_column : int
		method get_edge_mode : edge_mode
		method get_end_at_last_line : bool
		method get_end_styled : int
		method get_eol_mode : eol_mode
		method get_first_visible_line : int
		method get_focus : bool
		method get_fold_expanded : int -> bool
		method get_fold_level : int -> int
		method get_fold_parent : int -> int
		method get_h_scrollbar : bool
		method get_highlight_guide : int
		method get_indent : int
		method get_indentation_guides : bool
		method get_last_child : int -> int -> int
		method get_layout_cache : cache_mode
		method get_length : int
		method get_lexer : int
		method get_line : int -> string
		method get_line_count : int
		method get_line_end_position : int -> int
		method get_line_indentation : int -> int
		method get_line_indentation_position : int -> int
		method get_line_sel_end_position : int -> int
		method get_line_sel_start_position : int -> int
		method get_line_state : int -> int
		method get_line_visible : int -> int
		method get_margin_left : int
		method get_margin_mask_n : int32
		method get_margin_right : int
		method get_margin_sensitive_n : bool
		method get_margin_type_n : margin_type
		method get_margin_width_n : int
		method get_max_line_state : int
		method get_mod_event_mask : modification_type list
		method get_modify : int
		method get_mouse_down_captures : bool
		method get_mouse_dwell_time : int
		method get_overtype : bool
		method get_paste_convert_endings : bool
		method get_property : string -> string
		method get_property_expanded : string -> string
		method get_property_int : ?default:int -> string -> int
		method get_readonly : bool
		method get_scroll_width : int
		method get_search_flags : search_flag list
		method get_sel_alpha : int
		method get_sel_text : string
		method get_selection_end : int
		method get_selection_mode : selection_mode
		method get_selection_start : int
		method get_status : int
		method get_style_at : int -> int
		method get_style_bits : int
		method get_style_bits_needed : int
		method get_styled_text : text_range -> string
		method get_tab_indents : bool
		method get_tab_width : int
		method get_target_end : int
		method get_target_start : int
		method get_text : int -> string
		method get_text_length : int
		method get_text_range : text_range -> string
		method get_two_phase_draw : bool
		method get_undo_collection : bool
		method get_use_palette : bool
		method get_use_tabs : bool
		method get_v_scrollbar : bool
		method get_view_eol : bool
		method get_view_ws : white_space_mode
		method get_wrap_mode : wrap_mode
		method get_wrap_start_indent : int
		method get_wrap_visual_flags : visual_flag list
		method get_wrap_visual_flags_location : visual_flag_loc list
		method get_x_offset : int
		method get_zoom : int
		method goto_line : int -> unit
		method goto_pos : int -> unit
		method grab_focus : unit -> unit
		method hide_lines : int -> int -> unit
		method hide_selection : bool -> unit
		method indic_get_fore : int -> int
		method indic_get_style : int -> indicator_style
		method indic_set_fore : int -> int -> unit
		method indic_set_style : int -> indicator_style -> unit
		method insert_text : int -> string -> unit
		method line_from_position : int -> int
		method line_length : int -> int
		method line_scroll : int -> int -> unit
		method lines_join : unit -> unit
		method lines_on_screen : int
		method lines_split : int -> unit
		method load_lexer_library : string -> unit
		method marker_add : int -> int -> int
		method marker_add_set : int -> int32 -> unit
		method marker_define : int -> marker_symbol -> unit
		method marker_define_pixmap : int -> int -> unit
		method marker_delete : int -> int -> unit
		method marker_delete_all : int -> unit
		method marker_delete_handle : int -> unit
		method marker_get : int -> int32
		method marker_line_from_handle : int -> int
		method marker_next : int -> int32 -> int
		method marker_previous : int -> int32 -> int
		method marker_set_alpha : int -> int -> unit
		method marker_set_back : int -> int -> unit
		method marker_set_fore : int -> int -> unit
		method move_caret_inside_view : unit -> unit
		method null : unit
		method private obj : scintilla Gtk.obj
		method paste : unit -> unit
		method point_x_from_position : int -> int
		method point_y_from_position : int -> int
		method position_after : int -> int
		method position_before : int -> int
		method position_from_line : int -> int
		method position_from_point : int -> int -> int
		method position_from_point_close : int -> int -> int
		method redo : unit -> unit
		method register_image : int -> string -> unit
		method release_document : document -> unit
		method replace_sel : string -> unit
		method replace_target : int -> string -> int
		method replace_target_re : int -> string -> int
		method scroll_caret : unit -> unit
		method search_anchor : unit -> unit
		method search_in_target : int -> string -> int
		method search_next : ?flags:search_flag list -> string -> int
		method search_prev : ?flags:search_flag list -> string -> int
		method select_all : unit -> unit
		method selection_is_rectangle : int
		method send_key_cmd : key_cmd -> unit
		method set_anchor : int -> unit
		method set_backspace_unindents : bool -> unit
		method set_buffered_draw : bool -> unit
		method set_caret_fore : int -> unit
		method set_caret_line_back : int -> unit
		method set_caret_line_back_alpha : int -> unit
		method set_caret_line_visible : bool -> unit
		method set_caret_period : int -> unit
		method set_caret_width : int -> unit
		method set_chars_default : unit -> unit
		method set_code_page : int -> unit
		method set_control_char_symbol : bool -> unit
		method set_current_pos : int -> unit
		method set_cursor : int -> unit
		method set_doc_pointer : document -> unit
		method set_edge_colour : int -> unit
		method set_edge_column : int -> unit
		method set_edge_mode : edge_mode -> unit
		method set_end_at_last_line : bool -> unit
		method set_eol_mode : eol_mode -> unit
		method set_focus : bool -> unit
		method set_fold_expanded : int -> bool -> unit
		method set_fold_flags : int -> unit
		method set_fold_level : int -> int -> unit
		method set_fold_margin_colour : bool -> int -> unit
		method set_fold_margin_hi_colour : bool -> int -> unit
		method set_h_scrollbar : bool -> unit
		method set_highlight_guide : int -> unit
		method set_hotspot_active_back : bool -> int -> unit
		method set_hotspot_active_fore : bool -> int -> unit
		method set_hotspot_active_underline : bool -> unit
		method set_id : int -> unit
		method set_indent : int -> unit
		method set_indentation_guides : bool -> unit
		method set_keywords : int -> string -> unit
		method set_layout_cache : cache_mode -> unit
		method set_length_for_encode : int -> unit
		method set_lexer : int -> unit
		method set_lexer_language : string -> unit
		method set_line_indentation : int -> int -> unit
		method set_line_state : int -> int -> unit
		method set_margin_left : int -> unit
		method set_margin_mask_n : int -> int32 -> unit
		method set_margin_right : int -> unit
		method set_margin_sensitive_n : int -> bool -> unit
		method set_margin_type_n : int -> margin_type -> unit
		method set_margin_width_n : int -> int -> unit
		method set_mod_event_mask : ?flags:modification_type list -> unit -> unit
		method set_mouse_down_captures : bool -> unit
		method set_mouse_dwell_time : int -> unit
		method set_overtype : bool -> unit
		method set_paste_convert_endings : bool -> unit
		method set_property : string -> string -> unit
		method set_readonly : bool -> unit
		method set_save_point : unit -> unit
		method set_scroll_width : int -> unit
		method set_search_flags : ?flags:search_flag list -> unit -> unit
		method set_sel : int -> int -> unit
		method set_sel_alpha : int -> unit
		method set_sel_back : bool -> int -> unit
		method set_sel_fore : bool -> int -> unit
		method set_selection_end : int -> unit
		method set_selection_mode : selection_mode -> unit
		method set_selection_start : int -> unit
		method set_status : int -> unit
		method set_style_bits : int -> unit
		method set_styling : int -> int -> unit
		method set_styling_ex : int -> string -> unit
		method set_tab_indents : bool -> unit
		method set_tab_width : int -> unit
		method set_target_end : int -> unit
		method set_target_start : int -> unit
		method set_text : string -> unit
		method set_two_phase_draw : bool -> unit
		method set_undo_collection : bool -> unit
		method set_use_palette : bool -> unit
		method set_use_tabs : bool -> unit
		method set_v_scrollbar : bool -> unit
		method set_view_eol : bool -> unit
		method set_view_ws : white_space_mode -> unit
		method set_visible_policy : ?flags:visible_policy list -> int -> unit
		method set_white_space_back : bool -> int -> unit
		method set_white_space_fore : bool -> int -> unit
		method set_whitespace_chars : string -> unit
		method set_word_chars : string -> unit
		method set_wrap_mode : wrap_mode -> unit
		method set_wrap_start_indent : int -> unit
		method set_wrap_visual_flags : ?flags:visual_flag list -> unit -> unit
		method set_wrap_visual_flags_location :
			?flags:visual_flag_loc list -> unit -> unit
		method set_x_caret_policy : ?flags:caret_policy list -> int -> unit
		method set_x_offset : int -> unit
		method set_y_caret_policy : ?flags:caret_policy list -> int -> unit
		method set_zoom : int -> unit
		method show_lines : int -> int -> unit
		method start_record : unit -> unit
		method start_styling : int -> int -> unit
		method stop_record : unit -> unit
		method style_clear_all : unit -> unit
		method style_reset_default : unit -> unit
		method style_set_back : int -> int -> unit
		method style_set_bold : int -> bool -> unit
		method style_set_case : int -> case_mode -> unit
		method style_set_changeable : int -> bool -> unit
		method style_set_characterset : int -> char_set -> unit
		method style_set_eolfilled : int -> bool -> unit
		method style_set_font : int -> string -> unit
		method style_set_fore : int -> int -> unit
		method style_set_hotspot : int -> bool -> unit
		method style_set_italic : int -> bool -> unit
		method style_set_size : int -> int -> unit
		method style_set_underline : int -> bool -> unit
		method style_set_visible : int -> bool -> unit
		method target_as_utf8 : string
		method target_from_selection : unit -> unit
		method text_height : int -> int
		method text_width : int -> string -> int
		method toggle_fold : int -> unit
		method undo : unit -> unit
		method use_popup : bool -> unit
		method user_list_show : int -> string -> unit
		method visible_from_doc_line : int -> int
		method word_end_position : int -> bool -> int
		method word_start_position : int -> bool -> int
		method zoom_in : unit -> unit
		method zoom_out : unit -> unit
	end

(* get [GObject] type info *)
val get_type : Gobject.g_type

(* create a new editor object instance *)
val editor : ?packing:(GObj.widget -> 'a) -> unit -> editor

(* the mysterious "release resources" - use when all instances are done *)
val release_resources : unit -> unit
